local Utopia = Utopia
if (not Utopia) then
	return
end
Utopia:UpdateVersion("$Revision: 26 $")

local new, del, copy, deepDel = Utopia.new, Utopia.del, Utopia.copy, Utopia.deepDel
local Gradient, UnitFullName, ClassColour, SmoothColour = Utopia.Gradient, Utopia.UnitFullName, Utopia.ClassColour, Utopia.SmoothColour
local rotate = Utopia.rotate
local band = bit.band
local savageCombat = GetSpellInfo(58413)
local bloodFrenzy = GetSpellInfo(29859)

----------------- BUG FIX ----- Remove this when WoW includes Savage Combat/Blood Frenzy in combat log --------------
local SChaxCache = setmetatable({},{
	__index = function(self,key)
		local val, class = UnitClass(key)
		if (class == "ROGUE") then
			val = Utopia:HasTalent(key, GetSpellInfo(58413))	-- Savage Combat
		elseif (class == "WARRIOR") then
			val = Utopia:HasTalent(key, GetSpellInfo(29859))	-- Blood Frenzy
		end
		self[key] = val or false
		return val
	end
})
--------------END BUG FIX ----- Remove this when WoW includes Savage Combat in combat log --------------

local engine = CreateFrame("Frame", "UtopiaDebuffActivity")
--@debug@
E = engine
--@no-debug@

function engine:Print(...)
	ChatFrame1:AddMessage("|cFFFFFF80Engine: |r"..format(...))
end

engine:SetScript("OnEvent", function(self, event, ...)
	local f = engine[event]
	if (f) then
		f(self, ...)
	else
		error("No handler for "..event)
	end
end)

-- Start
function engine:Start()
	self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
	self:RegisterEvent("UNIT_FLAGS")
	self:RegisterEvent("RAID_ROSTER_UPDATE")
	self:RegisterEvent("PARTY_MEMBERS_CHANGED")
	self:RegisterEvent("UNIT_NAME_UPDATE")
	self:RegisterEvent("UNIT_PET")

	self.mobs = new()
	self.players = new()
	self:ScanRoster()
	self:ResetInCombatForAll()
end

-- Stop
function engine:Stop()
	self:UnregisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
	self:UnregisterEvent("UNIT_FLAGS")
	self:UnregisterEvent("RAID_ROSTER_UPDATE")
	self:UnregisterEvent("PARTY_MEMBERS_CHANGED")
	self:UnregisterEvent("UNIT_NAME_UPDATE")
	self:UnregisterEvent("UNIT_PET")
	self:UnregisterEvent("UNIT_AURA")

	deepDel(self.mobs)
	deepDel(self.players)
	deepDel(self.combat)
end

local function checkRosterUpdate(self, elapsed)
	if (self.rosterUpdate) then
		self.rosterUpdate = nil
		self:SetScript("OnUpdate", nil)
		self:ScanRoster()
	end
end

function engine:RAID_ROSTER_UPDATE()
	self.rosterUpdate = true
	self:SetScript("OnUpdate", checkRosterUpdate)
end
engine.PARTY_MEMBERS_CHANGED = engine.RAID_ROSTER_UPDATE

-- UNIT_NAME_UPDATE
function engine:UNIT_NAME_UPDATE(unit)
	local guid = UnitGUID(unit)
	local player = self.players[guid]
	if (player) then
		player.name = UnitFullName(unit)
	else
		local mob = self.mobs[guid]
		if (mob) then
			mob.name = UnitName(unit)
		end
	end
end

-- UNIT_PET
function engine:UNIT_PET(unit)
	if (not UnitInParty(unit) and not UnitInRaid(unit)) then
		return
	end

	local ownerGUID = UnitGUID(unit)
	if (not ownerGUID) then
		-- Probably zoning
		return 
	end

	local petid
	if (unit == "player") then
		petid = "pet"
	else
		petid = unit:gsub("^(%a+)(%d+)$", "%1pet%2")
	end

	local player = self.players[ownerGUID]
	assert(player, "player ~= nil")

	if (UnitExists(petid) and UnitCreatureFamily(petid)) then
		local petGUID = UnitGUID(petid)
		if (not self.players[petGUID]) then
			self:AddPlayer(petid, unit)
			player.pet = petGUID
			--self:Print("Added pet for %s called %s", UnitName(unit), UnitName(petid))
		end
	else
		local petGUID = player.pet
		if (petGUID) then
			player.pet = nil
			assert(self.players[petGUID])
			--self:Print("Removed pet from %s", UnitName(unit))
			self.players[petGUID] = del(self.players[petGUID])
		end
	end
end

-- AddPlayer
function engine:AddPlayer(unit, ownerid)
	local name = UnitFullName(unit)
	local guid = UnitGUID(unit)

	local n = self.oldRoster[guid]
	if (n) then
		self.oldRoster[guid] = nil
	else
		n = new()
	end
	self.players[guid] = n

	n.name = name
	n.guid = guid
	n.ownerid = ownerid

	return n
end

-- ScanRoster
function engine:ScanRoster()
	if (not self.players) then
		return
	end

	self.oldRoster = self.players
	self.players = new()
	self.scRogues = 0		-- BUG FIX ----- Remove this when WoW includes Savage Combat in combat log --------------
	self.bfWarriors = 0

	for unit in Utopia:IterateRoster() do
		local player = self:AddPlayer(unit)

		----------------- BUG FIX ----- Remove this when WoW includes Savage Combat in combat log --------------
		local _, class = UnitClass(unit)
		if (class == "ROGUE") then
			if (Utopia.talents and Utopia.talents[UnitFullName(unit)]) then
				if (Utopia:HasTalent(unit, GetSpellInfo(58413))) then
					self.scRogues = self.scRogues + 1
				end
			--else
			--	self.scRogues = self.scRogues + 1
			end
		elseif (class == "WARRIOR") then
			-- Dittoe for Warrior's Blood Frenzy
			if (Utopia.talents and Utopia.talents[UnitFullName(unit)]) then
				if (Utopia:HasTalent(unit, GetSpellInfo(29859))) then
					self.bfWarriors = self.bfWarriors + 1
				end
			--else
			--	self.scRogues = self.scRogues + 1
			end
		end
		-------------- END BUG FIX ----- Remove this when WoW includes Savage Combat in combat log -------------

		local petid
		if (unit == "player") then
			petid = "pet"
		else
			petid = unit:gsub("^(%a+)(%d+)$", "%1pet%2")
		end

		if (UnitExists(petid) and UnitCreatureFamily(petid)) then
			self:AddPlayer(petid, unit)
			player.pet = UnitGUID(petid)
		end
	end

	----------------- BUG FIX ----- Remove this when WoW includes Savage Combat in combat log --------------
	if (self.scRogues > 0 or self.bfWarriors > 0) then
		self:RegisterEvent("UNIT_AURA")
	else
		self:UnregisterEvent("UNIT_AURA")
	end
	-------------- END BUG FIX ----- Remove this when WoW includes Savage Combat in combat log -------------

	deepDel(self.oldRoster)
end

-- COMBAT_LOG_EVENT_UNFILTERED
local mask = COMBATLOG_OBJECT_AFFILIATION_PARTY + COMBATLOG_OBJECT_AFFILIATION_RAID + COMBATLOG_OBJECT_AFFILIATION_MINE
local dstMask = COMBATLOG_OBJECT_REACTION_HOSTILE + COMBATLOG_OBJECT_REACTION_NEUTRAL + COMBATLOG_OBJECT_CONTROL_NPC

function engine:COMBAT_LOG_EVENT_UNFILTERED(timestamp, event, ...)
	--for i = 1,select("#", ...) do
	--	local str = tostring(select(i, ...))
	--	if (str == "Savage Combat") then
	--		self:Print("|cFFFF0000%s in event %s", str, event)
	--		X = {timestamp, event, ...}
	--	end
	--end

	local ev = self.clEvents[event]
	if (ev) then
		ev(self, timestamp, event, ...)
	end
end

engine.clEvents = {}
-- SPELL_AURA_APPLIED
function engine.clEvents:SPELL_AURA_APPLIED(timestamp, event, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, spellId, spellName, spellSchool, auraType)
	if (band(srcFlags, mask) ~= 0) then
		if (band(dstFlags, dstMask) == 0) then
			return
		end

		----------------- BUG FIX ----- Remove this when WoW includes Savage Combat in combat log --------------
		if (self.scRogues > 0 and spellSchool == SCHOOL_MASK_NATURE) then
			-- When a rogue with Savage Combat talent applies a poison, we remember who
			-- they are and the target they hit for next UNIT_AURA check... YUK!
			local _, class = UnitClass(srcName)
			if (class == "ROGUE") then
				if (SChaxCache[srcName]) then				-- Talent check cached for speed
					if (not self[savageCombat]) then
						self[savageCombat] = new()
					end
					--self:Print("Flagging |cFFFF8080Savage Combat|r potential for |cFFFFFF80%s|r on |cFFFF0000%s", srcName, dstName)
					self[savageCombat][dstGUID] = srcGUID
					return
				end
			end
		elseif (self.bfWarriors > 0 and spellSchool == SCHOOL_MASK_PHYSICAL) then
			if (spellName == GetSpellInfo(772) or spellId == 12721) then			-- Rend & Deep Wounds
				-- Same for Warriors who apply Rend
				local _, class = UnitClass(srcName)
				if (class == "WARRIOR") then
					if (SChaxCache[srcName]) then				-- Talent check cached for speed
						if (not self[bloodFrenzy]) then
							self[bloodFrenzy] = new()
						end
						--self:Print("Flagging |cFFFF8080Blood Frenzy|r potential for |cFFFFFF80%s|r on |cFFFF0000%s", srcName, dstName)
						self[bloodFrenzy][dstGUID] = srcGUID
						return
					end
				end
			end
		end
		------------- END BUG FIX ----- Remove this when WoW includes Savage Combat in combat log --------------

		if (not Utopia.lookup[spellName]) then
			return
		end

		--self:Print("%s: |cFFFFFF80%s|r applies |cFFFF8080%s|r to |cFFFF0000%s|r - School = %s", event, srcName, spellName, dstName, tostring(spellSchool))
		local player = self.players[srcGUID]
		if (player) then
			local mob = self.mobs[dstGUID]
			if (not mob) then
				--self:Print("- New Mob: %s", dstName)
				mob = new()
				self.mobs[dstGUID] = mob

				mob.name = dstName
				mob.guid = dstGUID
				mob.debuffs = new()
			end

			local debuff = mob.debuffs[spellName]
			if (not debuff) then
				--self:Print("- New debuff for %s: %s", dstName, spellName)
				debuff = new()
				mob.debuffs[spellName] = debuff

				debuff.name = spellName
				debuff.id = spellId
				debuff.source = srcGUID
				debuff.stacks = 1
			else
				self:Print("SPELL_AURA_APPLIED for |cFFFF8080%s|r that already exists on |cFFFF0000%s|r", spellName, dstName)
			end

			debuff.startTime = GetTime()

			Utopia:OnDebuffApplied(dstGUID, spellName, debuff)
		else
			error("Passed mask test, but found no matching engine.player for "..srcName)
		end
	end
end

-- SPELL_AURA_APPLIED_DOSE
function engine.clEvents:SPELL_AURA_APPLIED_DOSE(timestamp, event, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, spellId, spellName, spellSchool, auraType)
	if (not Utopia.lookup[spellName]) then
		return
	end
	--self:Print("%s: |cFFFFFF80%s|r applies |cFFFF8080%s|r to |cFFFF0000%s|r", event, srcName, spellName, dstName)
	if (band(dstFlags, dstMask) == 0) then
		return
	end

	local mob = self.mobs[dstGUID]
	if (not mob) then
		-- self:Print("SPELL_AURA_APPLIED_DOSE(1) |cFFFF8080%s|r without a SPELL_AURA_APPLIED for %s", spellName or "nil", dstName or "nil")
		-- Probably targetted a mob being killed by someone else
		return
	end

	local debuff = mob.debuffs[spellName]
	if (not debuff) then
		self:Print("SPELL_AURA_APPLIED_DOSE(2) |cFFFF8080%s|r without a SPELL_AURA_APPLIED for %s", spellName or "nil", dstName or "nil")
	else
		debuff.startTime = GetTime()
		debuff.stacks = (debuff.stacks or 0) + 1
		--self:Print("- Stacks for |cFFFF0000%s's|r |cFFFF8080%s|r incremented to |cFF00FF00%d|r", dstName, spellName, debuff.stacks)

		Utopia:OnDebuffApplied(dstGUID, spellName, debuff, true)
	end
end


-- SPELL_AURA_REMOVED
function engine.clEvents:SPELL_AURA_REMOVED(timestamp, event, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, spellId, spellName, spellSchool, auraType)
	if (band(srcFlags, mask) ~= 0) then
		if (not Utopia.lookup[spellName]) then
			return
		end
		if (band(dstFlags, dstMask) == 0) then
			return
		end

		--self:Print("%s: |cFFFFFF80%s's|r debuff of |cFFFF8080%s|r removed from |cFFFF0000%s|r", event, srcName, spellName, dstName)
		local player = self.players[srcGUID]
		if (player) then
			local mob = self.mobs[dstGUID]
			if (mob) then
				local debuff = mob.debuffs[spellName]
				if (debuff) then
					Utopia:OnDebuffRemoved(dstGUID, spellName, debuff)
					mob.debuffs[spellName] = nil
				else
					self:Print("SPELL_AURA_REMOVED: with no matching mob debuff (%s)", spellName)
				end
			--else
				-- Probably ran out of combat and the mob's debuffs auto expired. This is ok.
				--self:Print("SPELL_AURA_REMOVED: with no matching mob (%s)", dstName)
			end
		else
			self:Print("SPELL_AURA_REMOVED: with no matching player (%s)", srcName)
		end
	end
end

-- SPELL_AURA_REMOVED_DOSE
function engine.clEvents:SPELL_AURA_REMOVED_DOSE(timestamp, event, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, spellId, spellName, spellSchool, auraType)
	if (not Utopia.lookup[spellName]) then
		return
	end
	if (band(dstFlags, dstMask) == 0) then
		return
	end

	local mob = self.mobs[dstGUID]
	if (mob) then
		local debuff = mob.debuffs[spellName]
		if (debuff) then
			debuff.stacks = debuff.stacks - 1
			Utopia:OnDebuffRemoved(dstGUID, spellName, debuff, true)
			if (debuff.stacks < 1) then
				self:Print("|cFFFF0000%s Stacks decremented to %d", spellName, debuff.stacks)
			end
		else
			self:Print("SPELL_AURA_REMOVED_DOSE: with no mob.debuffs['%s']", spellName)
		end
	else
		self:Print("SPELL_AURA_REMOVED_DOSE: with no mob (%s)", dstName)
	end
end

----------------- BUG FIX ----- Remove this when WoW includes Savage Combat in combat log --------------
function engine:ApplyHackedArua(unit, spellName, onoff, ...)
	local dstGUID = UnitGUID(unit)
	local srcGUID = self[spellName] and self[spellName][dstGUID]

	if (onoff) then
		if (not self.has) then
			self.has = new()
		end
		if (not self.has[spellName]) then
			self.has[spellName] = new()
		end

		if (not self.has[spellName][dstGUID]) then
			self.has[spellName][dstGUID] = srcGUID

			local player = self.players[srcGUID]
			if (player) then
				local mob = self.mobs[dstGUID]
				if (not mob) then
					mob = new()
					self.mobs[dstGUID] = mob
	
					mob.name = UnitName(unit)
					mob.guid = dstGUID
					mob.debuffs = new()
				end
	
				local debuff = mob.debuffs[spellName]
				if (not debuff) then
					debuff = new()
					mob.debuffs[spellName] = debuff
				end
	
				debuff.name = spellName
				debuff.id = select(onoff, ...)
				debuff.source = srcGUID
				debuff.stacks = 1
	
				Utopia:OnDebuffApplied(dstGUID, spellName, debuff)
			end
		end
	else
		local mob = self.mobs[dstGUID]
		if (mob) then
			local debuff = mob.debuffs[spellName]
			if (debuff) then
				if (self.has and self.has[spellName] and self.has[spellName][dstGUID]) then
					self.has[spellName][dstGUID] = nil
--@debug@
				else
					self:Print("Removing %s from %s but we weren't supposed to have it", spellName, UnitName(unit))
--@no-debug@
				end

				Utopia:OnDebuffRemoved(dstGUID, spellName, debuff)
				mob.debuffs[spellName] = nil
			end
		end
	end
end

-- UNIT_AURA
function engine:UNIT_AURA(unit)
	if (not UnitCanAttack("player", unit)) then
		return
	end

	local foundSC, foundBF
	for i = 1,100 do
		local name, rank, tex = UnitDebuff(unit, i)
		if (name == nil) then
			break
		elseif (name == savageCombat) then
			foundSC = tonumber(rank:match("(%d)")) or 2
		elseif (name == bloodFrenzy) then
			foundBF = tonumber(rank:match("(%d)")) or 2
		end
	end

	if (self.scRogues > 0) then
		self:ApplyHackedArua(unit, savageCombat, foundSC, 51682, 58413)
	end
	if (self.bfWarriors > 0) then
		self:ApplyHackedArua(unit, bloodFrenzy, foundBF, 29836, 29859)
	end
end
------------- END BUG FIX ----- Remove this when WoW includes Savage Combat in combat log --------------

-- ResetInCombatForAll
function engine:ResetInCombatForAll()
	del(self.combat)
	self.combat = new()
	for unit in Utopia:IterateRoster() do
		self.combat[UnitGUID(unit)] = UnitAffectingCombat(unit)
	end
end

-- UNIT_FLAGS
function engine:UNIT_FLAGS(unit)
	local guid = UnitGUID(unit)
	if (UnitAffectingCombat(unit)) then
		if (UnitInParty(unit) or UnitPlayerOrPetInRaid(unit) or UnitPlayerOrPetInParty(unit)) then
			if (not self.combat[guid]) then
				self.combat[guid] = true
			end
		end
	else
		if (self.combat[guid]) then
			self.combat[guid] = nil

			if (not next(self.combat) or UnitIsUnit(unit, "player")) then
				-- Nothing flagged as in-combat, but we double check to be sure
				-- for cases where ppl become IsVisible() and we don't see a
				-- UNIT_FLAGS event for them. Also checking this when I go out
				-- of combat to be doubly sure things don't go adrift over time
				self:ResetInCombatForAll()
				if (not next(self.combat)) then
					self:Cleanup()
				end
			end
		end
	end
end

-- Cleanup
function engine:Cleanup()
	deepDel(self.mobs)
	self.mobs = new()

	----------------- BUG FIX ----- Remove this when WoW includes Savage Combat/Blood Frenzy in combat log --------------
	self[savageCombat] = del(self[savageCombat])
	self[bloodFrenzy] = del(self[bloodFrenzy])
	self.has = deepDel(self.has)
	for k in pairs(SChaxCache) do
		SChaxCache[k] = nil
	end
	------------- END BUG FIX ----- Remove this when WoW includes Savage Combat in combat log --------------
end

-- GetMobByUnit
function engine:GetMobByUnit(unit)
	return engine.mobs and engine.mobs[UnitGUID(unit)]
end

-- ShowFunc
-- Debug dump of all current debuffs on all targets
function engine:ShowFunc()
	if (self.mobs) then
		for guid, mob in pairs(self.mobs) do
			self:Print("|cFFFF0000%s|cFF808080 (%s)", mob.name, guid)

			if (next(mob.debuffs)) then
				for debuffName, debuff in pairs(mob.debuffs) do
					local playerGUID = debuff.source
					local player = engine.players[playerGUID]

					if (player) then
						self:Print("- |cFFFFFF80%s|r applied: |cFFFF8080%s|r (id=%d) (|cFF80FF80%s|r stacks)", player.name, debuffName, debuff.id, debuff.stacks)
					else
						self:Print("- |cFFFF00FFUnknown|r applied %s (id=%d) (|cFF80FF80%s|r stacks)", debuffName, debuff.id, debuff.stacks)
					end
				end
			else
				self:Print("- No debuffs")
			end
		end
	end
end

local function UE()
	engine:ShowFunc()
end

SlashCmdList["UE"] = UE
SLASH_UE1 = "/ue"
